[Unit]
Description={{ description }}
Documentation=man:{{ name }}.8
After=network.target

[Service]
WorkingDirectory=/etc/sv/{{ name }}
ExecStart=/usr/bin/chpst -e conf /bin/sh run
SyslogIdentifier={{ name }}

LockPersonality=true
MemoryDenyWriteExecute=true
ProtectClock=true
ProtectControlGroups=true
ProtectHome=true
ProtectHostname=true
ProtectSystem=strict
PrivateDevices=true
RestrictAddressFamilies=AF_INET AF_INET6
RestrictNamespaces=true

[Install]
WantedBy=multi-user.target
